@@ -57,8 +57,9 @@ extern char **environ;
 
 #define GDM_FACTORY_SLAVE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GDM_TYPE_FACTORY_SLAVE, GdmFactorySlavePrivate))
 
-#define GDM_DBUS_NAME                      "org.gnome.DisplayManager"
-#define GDM_DBUS_FACTORY_DISPLAY_INTERFACE "org.gnome.DisplayManager.StaticFactoryDisplay"
+#define GDM_DBUS_NAME                            "org.gnome.DisplayManager"
+#define GDM_DBUS_LOCAL_DISPLAY_FACTORY_PATH      "/org/gnome/DisplayManager/LocalDisplayFactory"
+#define GDM_DBUS_LOCAL_DISPLAY_FACTORY_INTERFACE "org.gnome.DisplayManager.LocalDisplayFactory"
 
 #define MAX_CONNECT_ATTEMPTS 10
 
@@ -66,8 +67,6 @@ struct GdmFactorySlavePrivate
 {
         char              *id;
         GPid               pid;
-        guint              output_watch_id;
-        guint              error_watch_id;
         guint              greeter_reset_id;
 
         GPid               server_pid;
@@ -78,7 +77,7 @@ struct GdmFactorySlavePrivate
         GdmSessionRelay   *session;
         GdmGreeterServer  *greeter_server;
         GdmGreeterSession *greeter;
-        DBusGProxy        *factory_display_proxy;
+        DBusGProxy        *factory_proxy;
         DBusGConnection   *connection;
 };
 
@@ -92,20 +91,57 @@ static void     gdm_factory_slave_finalize      (GObject             *object);
 
 G_DEFINE_TYPE (GdmFactorySlave, gdm_factory_slave, GDM_TYPE_SLAVE)
 
+static gboolean
+greeter_reset_timeout (GdmFactorySlave *slave)
+{
+        gdm_greeter_server_reset (slave->priv->greeter_server);
+        slave->priv->greeter_reset_id = 0;
+        return FALSE;
+}
+
 static void
-on_greeter_start (GdmGreeterSession *greeter,
-                  GdmFactorySlave   *slave)
+queue_greeter_reset (GdmFactorySlave *slave)
+{
+        if (slave->priv->greeter_reset_id > 0) {
+                return;
+        }
+
+        slave->priv->greeter_reset_id = g_timeout_add_seconds (2, (GSourceFunc)greeter_reset_timeout, slave);
+}
+
+static void
+on_greeter_session_start (GdmGreeterSession *greeter,
+                          GdmFactorySlave   *slave)
 {
         g_debug ("GdmFactorySlave: Greeter started");
 }
 
 static void
-on_greeter_stop (GdmGreeterSession *greeter,
-                 GdmFactorySlave   *slave)
+on_greeter_session_stop (GdmGreeterSession *greeter,
+                         GdmFactorySlave   *slave)
 {
         g_debug ("GdmFactorySlave: Greeter stopped");
 }
 
+static void
+on_greeter_session_exited (GdmGreeterSession    *greeter,
+                           int                   code,
+                           GdmFactorySlave      *slave)
+{
+        g_debug ("GdmSimpleSlave: Greeter exited: %d", code);
+        gdm_slave_stopped (GDM_SLAVE (slave));
+}
+
+static void
+on_greeter_session_died (GdmGreeterSession    *greeter,
+                         int                   signal,
+                         GdmFactorySlave      *slave)
+{
+        g_debug ("GdmSimpleSlave: Greeter died: %d", signal);
+        gdm_slave_stopped (GDM_SLAVE (slave));
+}
+
+
 static void
 on_session_info (GdmSession      *session,
                  const char      *text,
@@ -144,52 +180,126 @@ on_session_secret_info_query (GdmSession      *session,
 }
 
 static void
-on_session_opened (GdmSession      *session,
-                   GdmFactorySlave *slave)
+on_session_conversation_started (GdmSession      *session,
+                                 GdmFactorySlave *slave)
 {
-        g_debug ("GdmFactorySlave:  session opened");
+        g_debug ("GdmFactorySlave: session conversation started");
 
         gdm_greeter_server_ready (slave->priv->greeter_server);
 }
 
 static void
-on_session_user_verified (GdmSession      *session,
+on_session_setup_complete (GdmSession      *session,
+                           GdmFactorySlave *slave)
+{
+        gdm_session_authenticate (session);
+}
+
+static void
+on_session_setup_failed (GdmSession      *session,
+                         const char      *message,
+                         GdmFactorySlave *slave)
+{
+        gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to initialize login system"));
+
+        queue_greeter_reset (slave);
+}
+
+static void
+on_session_reset_complete (GdmSession      *session,
+                           GdmFactorySlave *slave)
+{
+        g_debug ("GdmFactorySlave: PAM reset");
+}
+
+static void
+on_session_reset_failed (GdmSession      *session,
+                         const char      *message,
+                         GdmFactorySlave *slave)
+{
+        g_critical ("Unable to reset PAM");
+}
+
+static void
+on_session_authenticated (GdmSession      *session,
                           GdmFactorySlave *slave)
+{
+        gdm_session_authorize (session);
+}
+
+static void
+on_session_authentication_failed (GdmSession      *session,
+                                  const char      *message,
+                                  GdmFactorySlave *slave)
+{
+        gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to authenticate user"));
+
+        queue_greeter_reset (slave);
+}
+
+static void
+on_session_authorized (GdmSession      *session,
+                       GdmFactorySlave *slave)
+{
+        int flag;
+
+        /* FIXME: check for migration? */
+        flag = GDM_SESSION_CRED_ESTABLISH;
+
+        gdm_session_accredit (session, flag);
+}
+
+static void
+on_session_authorization_failed (GdmSession      *session,
+                                 const char      *message,
+                                 GdmFactorySlave *slave)
+{
+        gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to authorize user"));
+
+        queue_greeter_reset (slave);
+}
+
+static void
+on_session_accredited (GdmSession      *session,
+                       GdmFactorySlave *slave)
 {
         g_debug ("GdmFactorySlave:  session user verified");
 
+        gdm_session_open_session (session);
+}
+
+static void
+on_session_accreditation_failed (GdmSession      *session,
+                                 const char      *message,
+                                 GdmFactorySlave *slave)
+{
+        g_debug ("GdmFactorySlave: could not successfully authenticate user: %s",
+                 message);
+
+        gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to establish credentials"));
+
+        queue_greeter_reset (slave);
+}
+
+static void
+on_session_opened (GdmSession      *session,
+                   GdmFactorySlave *slave)
+{
+        g_debug ("GdmFactorySlave: session opened");
+
         gdm_session_start_session (session);
 
         gdm_greeter_server_reset (slave->priv->greeter_server);
 }
 
-static gboolean
-greeter_reset_timeout (GdmFactorySlave *slave)
-{
-        gdm_greeter_server_reset (slave->priv->greeter_server);
-        slave->priv->greeter_reset_id = 0;
-        return FALSE;
-}
-
 static void
-queue_greeter_reset (GdmFactorySlave *slave)
+on_session_open_failed (GdmSession      *session,
+                        const char      *message,
+                        GdmFactorySlave *slave)
 {
-        if (slave->priv->greeter_reset_id > 0) {
-                return;
-        }
+        g_debug ("GdmFactorySlave: could not open session: %s", message);
 
-        slave->priv->greeter_reset_id = g_timeout_add_seconds (2, (GSourceFunc)greeter_reset_timeout, slave);
-}
-
-static void
-on_session_user_verification_error (GdmSession      *session,
-                                    const char      *message,
-                                    GdmFactorySlave *slave)
-{
-        g_debug ("GdmFactorySlave: could not successfully authenticate user: %s",
-                 message);
-
-        gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to authenticate user"));
+        gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to open session"));
 
         queue_greeter_reset (slave);
 }
@@ -199,13 +309,14 @@ on_session_session_started (GdmSession      *session,
                             GdmFactorySlave *slave)
 {
         g_debug ("GdmFactorySlave: Relay session started");
+
         gdm_greeter_server_reset (slave->priv->greeter_server);
 }
 
 static gboolean
 create_product_display (GdmFactorySlave *slave)
 {
-        char    *display_id;
+        char    *parent_display_id;
         char    *server_address;
         char    *product_id;
         GError  *error;
@@ -216,33 +327,33 @@ create_product_display (GdmFactorySlave *slave)
 
         g_debug ("GdmFactorySlave: Create product display");
 
-        g_object_get (slave,
-                      "display-id", &display_id,
-                      NULL);
-
-        g_debug ("GdmFactorySlave: Connecting to display %s", display_id);
-        slave->priv->factory_display_proxy = dbus_g_proxy_new_for_name (slave->priv->connection,
-                                                                        GDM_DBUS_NAME,
-                                                                        display_id,
-                                                                        GDM_DBUS_FACTORY_DISPLAY_INTERFACE);
-        g_free (display_id);
-
-        if (slave->priv->factory_display_proxy == NULL) {
-                g_warning ("Failed to create display proxy %s", display_id);
+        g_debug ("GdmFactorySlave: Connecting to local display factory");
+        slave->priv->factory_proxy = dbus_g_proxy_new_for_name (slave->priv->connection,
+                                                                GDM_DBUS_NAME,
+                                                                GDM_DBUS_LOCAL_DISPLAY_FACTORY_PATH,
+                                                                GDM_DBUS_LOCAL_DISPLAY_FACTORY_INTERFACE);
+        if (slave->priv->factory_proxy == NULL) {
+                g_warning ("Failed to create local display factory proxy");
                 goto out;
         }
 
         server_address = gdm_session_relay_get_address (slave->priv->session);
 
+        g_object_get (slave,
+                      "display-id", &parent_display_id,
+                      NULL);
+
         error = NULL;
-        res = dbus_g_proxy_call (slave->priv->factory_display_proxy,
+        res = dbus_g_proxy_call (slave->priv->factory_proxy,
                                  "CreateProductDisplay",
                                  &error,
+                                 DBUS_TYPE_G_OBJECT_PATH, parent_display_id,
                                  G_TYPE_STRING, server_address,
                                  G_TYPE_INVALID,
                                  DBUS_TYPE_G_OBJECT_PATH, &product_id,
                                  G_TYPE_INVALID);
         g_free (server_address);
+        g_free (parent_display_id);
 
         if (! res) {
                 if (error != NULL) {
@@ -277,7 +388,7 @@ on_session_relay_connected (GdmSessionRelay *session,
 {
         g_debug ("GdmFactorySlave: Relay Connected");
 
-        gdm_session_open (GDM_SESSION (slave->priv->session));
+        gdm_session_start_conversation (GDM_SESSION (slave->priv->session));
 }
 
 static void
@@ -285,8 +396,8 @@ on_greeter_begin_verification (GdmGreeterServer *greeter_server,
                                GdmFactorySlave  *slave)
 {
         g_debug ("GdmFactorySlave: begin verification");
-
-        gdm_session_begin_verification (GDM_SESSION (slave->priv->session));
+        gdm_session_setup (GDM_SESSION (slave->priv->session),
+                           "gdm");
 }
 
 static void
@@ -295,9 +406,9 @@ on_greeter_begin_verification_for_user (GdmGreeterServer *greeter_server,
                                         GdmFactorySlave  *slave)
 {
         g_debug ("GdmFactorySlave: begin verification for user");
-
-        gdm_session_begin_verification_for_user (GDM_SESSION (slave->priv->session),
-                                                 username);
+        gdm_session_setup_for_user (GDM_SESSION (slave->priv->session),
+                                    "gdm",
+                                    username);
 }
 
 static void
@@ -325,6 +436,14 @@ on_greeter_language_selected (GdmGreeterServer *greeter_server,
         gdm_session_select_language (GDM_SESSION (slave->priv->session), text);
 }
 
+static void
+on_greeter_layout_selected (GdmGreeterServer *greeter_server,
+                            const char       *text,
+                            GdmFactorySlave  *slave)
+{
+        gdm_session_select_layout (GDM_SESSION (slave->priv->session), text);
+}
+
 static void
 on_greeter_user_selected (GdmGreeterServer *greeter_server,
                           const char       *text,
@@ -391,7 +510,7 @@ run_greeter (GdmFactorySlave *slave)
         /* FIXME: send a signal back to the master */
 
         /* Run the init script. gdmslave suspends until script has terminated */
-        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/Init", "gdm");
+        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/Init", GDM_USERNAME);
 
         slave->priv->greeter_server = gdm_greeter_server_new (display_id);
         g_signal_connect (slave->priv->greeter_server,
@@ -414,6 +533,10 @@ run_greeter (GdmFactorySlave *slave)
                           "language-selected",
                           G_CALLBACK (on_greeter_language_selected),
                           slave);
+        g_signal_connect (slave->priv->greeter_server,
+                          "layout-selected",
+                          G_CALLBACK (on_greeter_layout_selected),
+                          slave);
         g_signal_connect (slave->priv->greeter_server,
                           "user-selected",
                           G_CALLBACK (on_greeter_user_selected),
@@ -437,17 +560,25 @@ run_greeter (GdmFactorySlave *slave)
                                                         display_is_local);
         g_signal_connect (slave->priv->greeter,
                           "started",
-                          G_CALLBACK (on_greeter_start),
+                          G_CALLBACK (on_greeter_session_start),
                           slave);
         g_signal_connect (slave->priv->greeter,
                           "stopped",
-                          G_CALLBACK (on_greeter_stop),
+                          G_CALLBACK (on_greeter_session_stop),
+                          slave);
+        g_signal_connect (slave->priv->greeter,
+                          "exited",
+                          G_CALLBACK (on_greeter_session_exited),
+                          slave);
+        g_signal_connect (slave->priv->greeter,
+                          "died",
+                          G_CALLBACK (on_greeter_session_died),
                           slave);
         g_object_set (slave->priv->greeter,
                       "x11-authority-file", auth_file,
                       NULL);
-        gdm_greeter_session_set_server_address (slave->priv->greeter, address);
-        gdm_greeter_session_start (slave->priv->greeter);
+        gdm_welcome_session_set_server_address (GDM_WELCOME_SESSION (slave->priv->greeter), address);
+        gdm_welcome_session_start (GDM_WELCOME_SESSION (slave->priv->greeter));
 
         g_free (address);
 
@@ -478,29 +609,54 @@ idle_connect_to_display (GdmFactorySlave *slave)
                         g_warning ("Unable to connect to display after %d tries - bailing out", slave->priv->connection_attempts);
                         exit (1);
                 }
+                return TRUE;
         }
 
         return FALSE;
 }
 
 static void
-server_ready_cb (GdmServer *server,
-                 GdmFactorySlave  *slave)
+on_server_ready (GdmServer       *server,
+                 GdmFactorySlave *slave)
 {
         g_debug ("GdmFactorySlave: Server ready");
 
         g_timeout_add (500, (GSourceFunc)idle_connect_to_display, slave);
 }
 
+static void
+on_server_exited (GdmServer       *server,
+                  int              exit_code,
+                  GdmFactorySlave *slave)
+{
+        g_debug ("GdmFactorySlave: server exited with code %d\n", exit_code);
+
+        gdm_slave_stopped (GDM_SLAVE (slave));
+}
+
+static void
+on_server_died (GdmServer       *server,
+                int              signal_number,
+                GdmFactorySlave *slave)
+{
+        g_debug ("GdmFactorySlave: server died with signal %d, (%s)",
+                 signal_number,
+                 g_strsignal (signal_number));
+
+        gdm_slave_stopped (GDM_SLAVE (slave));
+}
+
 static gboolean
 gdm_factory_slave_run (GdmFactorySlave *slave)
 {
         char    *display_name;
+        char    *auth_file;
         gboolean display_is_local;
 
         g_object_get (slave,
                       "display-is-local", &display_is_local,
                       "display-name", &display_name,
+                      "display-x11-authority-file", &auth_file,
                       NULL);
 
         /* if this is local display start a server if one doesn't
@@ -508,11 +664,18 @@ gdm_factory_slave_run (GdmFactorySlave *slave)
         if (display_is_local) {
                 gboolean res;
 
-                slave->priv->server = gdm_server_new (display_name);
-
+                slave->priv->server = gdm_server_new (display_name, auth_file);
+                g_signal_connect (slave->priv->server,
+                                  "exited",
+                                  G_CALLBACK (on_server_exited),
+                                  slave);
+                g_signal_connect (slave->priv->server,
+                                  "died",
+                                  G_CALLBACK (on_server_died),
+                                  slave);
                 g_signal_connect (slave->priv->server,
                                   "ready",
-                                  G_CALLBACK (server_ready_cb),
+                                  G_CALLBACK (on_server_ready),
                                   slave);
 
                 res = gdm_server_start (slave->priv->server);
@@ -534,6 +697,7 @@ gdm_factory_slave_run (GdmFactorySlave *slave)
         }
 
         g_free (display_name);
+        g_free (auth_file);
 
         return TRUE;
 }
@@ -541,16 +705,67 @@ gdm_factory_slave_run (GdmFactorySlave *slave)
 static gboolean
 gdm_factory_slave_start (GdmSlave *slave)
 {
-        gboolean res;
         gboolean ret;
 
         ret = FALSE;
 
         g_debug ("GdmFactorySlave: Starting factory slave");
 
-        res = GDM_SLAVE_CLASS (gdm_factory_slave_parent_class)->start (slave);
+        GDM_SLAVE_CLASS (gdm_factory_slave_parent_class)->start (slave);
 
         GDM_FACTORY_SLAVE (slave)->priv->session = gdm_session_relay_new ();
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "conversation-started",
+                          G_CALLBACK (on_session_conversation_started),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "setup-complete",
+                          G_CALLBACK (on_session_setup_complete),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "setup-failed",
+                          G_CALLBACK (on_session_setup_failed),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "reset-complete",
+                          G_CALLBACK (on_session_reset_complete),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "reset-failed",
+                          G_CALLBACK (on_session_reset_failed),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "authenticated",
+                          G_CALLBACK (on_session_authenticated),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "authentication-failed",
+                          G_CALLBACK (on_session_authentication_failed),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "authorized",
+                          G_CALLBACK (on_session_authorized),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "authorization-failed",
+                          G_CALLBACK (on_session_authorization_failed),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "accredited",
+                          G_CALLBACK (on_session_accredited),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "accreditation-failed",
+                          G_CALLBACK (on_session_accreditation_failed),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "session-opened",
+                          G_CALLBACK (on_session_opened),
+                          slave);
+        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
+                          "session-open-failed",
+                          G_CALLBACK (on_session_open_failed),
+                          slave);
         g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
                           "info",
                           G_CALLBACK (on_session_info),
@@ -567,18 +782,6 @@ gdm_factory_slave_start (GdmSlave *slave)
                           "secret-info-query",
                           G_CALLBACK (on_session_secret_info_query),
                           slave);
-        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
-                          "user-verified",
-                          G_CALLBACK (on_session_user_verified),
-                          slave);
-        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
-                          "user-verification-error",
-                          G_CALLBACK (on_session_user_verification_error),
-                          slave);
-        g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
-                          "opened",
-                          G_CALLBACK (on_session_opened),
-                          slave);
         g_signal_connect (GDM_FACTORY_SLAVE (slave)->priv->session,
                           "session-started",
                           G_CALLBACK (on_session_session_started),
@@ -605,11 +808,9 @@ gdm_factory_slave_start (GdmSlave *slave)
 static gboolean
 gdm_factory_slave_stop (GdmSlave *slave)
 {
-        gboolean res;
-
         g_debug ("GdmFactorySlave: Stopping factory_slave");
 
-        res = GDM_SLAVE_CLASS (gdm_factory_slave_parent_class)->stop (slave);
+        GDM_SLAVE_CLASS (gdm_factory_slave_parent_class)->stop (slave);
 
         if (GDM_FACTORY_SLAVE (slave)->priv->session != NULL) {
                 gdm_session_relay_stop (GDM_FACTORY_SLAVE (slave)->priv->session);
@@ -624,7 +825,7 @@ gdm_factory_slave_stop (GdmSlave *slave)
         }
 
         if (GDM_FACTORY_SLAVE (slave)->priv->greeter != NULL) {
-                gdm_greeter_session_stop (GDM_FACTORY_SLAVE (slave)->priv->greeter);
+                gdm_welcome_session_stop (GDM_WELCOME_SESSION (GDM_FACTORY_SLAVE (slave)->priv->greeter));
                 g_object_unref (GDM_FACTORY_SLAVE (slave)->priv->greeter);
                 GDM_FACTORY_SLAVE (slave)->priv->greeter = NULL;
         }
@@ -635,8 +836,8 @@ gdm_factory_slave_stop (GdmSlave *slave)
                 GDM_FACTORY_SLAVE (slave)->priv->server = NULL;
         }
 
-        if (GDM_FACTORY_SLAVE (slave)->priv->factory_display_proxy != NULL) {
-                g_object_unref (GDM_FACTORY_SLAVE (slave)->priv->factory_display_proxy);
+        if (GDM_FACTORY_SLAVE (slave)->priv->factory_proxy != NULL) {
+                g_object_unref (GDM_FACTORY_SLAVE (slave)->priv->factory_proxy);
         }
 
         return TRUE;
@@ -648,10 +849,6 @@ gdm_factory_slave_set_property (GObject      *object,
                                const GValue *value,
                                GParamSpec   *pspec)
 {
-        GdmFactorySlave *self;
-
-        self = GDM_FACTORY_SLAVE (object);
-
         switch (prop_id) {
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -665,10 +862,6 @@ gdm_factory_slave_get_property (GObject    *object,
                                 GValue     *value,
                                 GParamSpec *pspec)
 {
-        GdmFactorySlave *self;
-
-        self = GDM_FACTORY_SLAVE (object);
-
         switch (prop_id) {
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -682,9 +875,6 @@ gdm_factory_slave_constructor (GType                  type,
                                GObjectConstructParam *construct_properties)
 {
         GdmFactorySlave      *factory_slave;
-        GdmFactorySlaveClass *klass;
-
-        klass = GDM_FACTORY_SLAVE_CLASS (g_type_class_peek (GDM_TYPE_FACTORY_SLAVE));
 
         factory_slave = GDM_FACTORY_SLAVE (G_OBJECT_CLASS (gdm_factory_slave_parent_class)->constructor (type,
                                                                                                          n_construct_properties,
